from django.contrib import admin
from django import forms
from django.utils.html import format_html
from django.urls import reverse
from dgenome.models import Chromosome, Gene, GeneLocation, GeneAlias
from dgenome.models import Transcript, Feature, Genome


class GenomeAdminForm(forms.ModelForm):
    class Meta:
        model = Genome
        fields = '__all__'


class GenomeAdmin(admin.ModelAdmin):
    form = GenomeAdminForm
    list_display = ['organism', 'chromosomes', 'assembly', 'source']
    readonly_fields = ['organism', 'chromosomes', 'assembly', 'source']


admin.site.register(Genome, GenomeAdmin)


class ChromosomeAdminForm(forms.ModelForm):
    class Meta:
        model = Chromosome
        fields = '__all__'


class ChromosomeAdmin(admin.ModelAdmin):
    form = ChromosomeAdminForm
    list_display = ['name', 'genome']
    readonly_fields = ['name', 'genome']

    def genome(self, obj):
        link = reverse('admin:dgenome_genome_change', args=[obj.genome.pk])
        return format_html('<a href="{}">{}</a>', link, obj.genome) if obj.genome else None


admin.site.register(Chromosome, ChromosomeAdmin)


class GeneAdminForm(forms.ModelForm):
    class Meta:
        model = Gene
        fields = '__all__'


class GeneAdmin(admin.ModelAdmin):
    form = GeneAdminForm
    list_display = ['name', 'description', 'Chr']
    readonly_fields = ['name', 'description', 'Chr']

    def Chr(self, obj):
        link = reverse('admin:dgenome_chromosome_change', args=[obj.chromosome.pk])
        return format_html('<a href="{}">{}</a>', link, obj.chromosome) if obj.chromosome else None


admin.site.register(Gene, GeneAdmin)


class GeneLocationAdminForm(forms.ModelForm):
    class Meta:
        model = GeneLocation
        fields = '__all__'


class GeneLocationAdmin(admin.ModelAdmin):
    form = GeneLocationAdminForm
    list_display = ['txstart', 'txend', 'cdsstart', 'cdsend', 'name']
    readonly_fields = ['txstart', 'txend', 'cdsstart', 'cdsend', 'name']

    def name(self, obj):
        link = reverse('admin:dgenome_gene_change', args=[obj.gene.pk])
        return format_html('<a href="{}">{}</a>', link, obj.gene) if obj.gene else None


admin.site.register(GeneLocation, GeneLocationAdmin)


class GeneAliasAdminForm(forms.ModelForm):
    class Meta:
        model = GeneAlias
        fields = '__all__'


class GeneAliasAdmin(admin.ModelAdmin):
    form = GeneAliasAdminForm
    list_display = ['alias']
    readonly_fields = ['alias']


admin.site.register(GeneAlias, GeneAliasAdmin)


class TranscriptAdminForm(forms.ModelForm):
    class Meta:
        model = Transcript
        fields = '__all__'


class TranscriptAdmin(admin.ModelAdmin):
    form = TranscriptAdminForm
    list_display = ['name', 'tss1500start', 'tss200start', 'txstart', 'txend',
                    'cdsstart', 'cdsend', 'description', 'long_name', 'gene_link']
    readonly_fields = ['name', 'tss1500start', 'tss200start', 'txstart', 'txend',
                       'cdsstart', 'cdsend', 'description', 'long_name', 'gene_link']

    def gene_link(self, obj):
        link = reverse('admin:dgenome_gene_change', args=[obj.gene.pk])
        return format_html('<a href="{}">{}</a>', link, obj.gene) if obj.gene else None


admin.site.register(Transcript, TranscriptAdmin)


class FeatureAdminForm(forms.ModelForm):
    class Meta:
        model = Feature
        fields = '__all__'


class FeatureAdmin(admin.ModelAdmin):
    form = FeatureAdminForm
    list_display = ['feat_type', 'start', 'end', 'trans_link']
    readonly_fields = ['feat_type', 'start', 'end', 'trans_link']

    def trans_link(self, obj):
        link = reverse('admin:dgenome_transcript_change', args=[obj.transcript.pk])
        return format_html('<a href="{}">{}</a>', link, obj.transcript) if obj.transcript else None


admin.site.register(Feature, FeatureAdmin)



from django.db import models


class Genome(models.Model):
    # Fields
    organism = models.CharField(max_length=100)
    chromosomes = models.IntegerField()
    assembly = models.CharField(max_length=100)
    source = models.CharField(max_length=100)

    class Meta:
        ordering = ('-pk',)

    def __str__(self):
        return '%s' % (self.assembly)

    def __unicode__(self):
        return u'%s' % self.pk


class Chromosome(models.Model):
    name = models.CharField(max_length=5, unique=True)

    # Relationship Fields
    genome = models.ForeignKey(
        Genome, on_delete=models.CASCADE
    )

    def __str__(self):
        return '%s' % (self.name)

    class Meta:
        ordering = ('-pk',)
        indexes = [
            models.Index(fields=['name']),
        ]

    def __unicode__(self):
        return u'%s' % self.pk


class Gene(models.Model):
    name = models.CharField(max_length=45)
    strand = models.CharField(max_length=50)
    description = models.TextField(max_length=3000)
    chromosome = models.ForeignKey(Chromosome, models.CASCADE)

    def __str__(self):
        return '%s' % (self.name)

    class Meta:
        ordering = ('-pk',)
        unique_together = ("name", "chromosome")
        indexes = [
            models.Index(fields=['name']),
        ]

    def __unicode__(self):
        return u'%s' % self.pk


class GeneLocation(models.Model):
    txstart = models.IntegerField()
    txend = models.IntegerField()
    cdsstart = models.IntegerField()
    cdsend = models.IntegerField()
    gene = models.ForeignKey(Gene, models.CASCADE)

    class Meta:
        ordering = ['txstart']
        unique_together = ("txstart", "txend", "gene")
        index_together = [
            ["txstart", "txend"],
        ]

    def __unicode__(self):
        return u'%s' % self.pk

    def get_gene_id(self):
        return self.gene.name


class GeneAlias(models.Model):
    alias = models.CharField(max_length=50)

    # Relationship Fields
    gene = models.ForeignKey(
        Gene, on_delete=models.CASCADE
    )

    class Meta:
        ordering = ('-pk',)
        indexes = [
            models.Index(fields=['alias']),
        ]

    def __unicode__(self):
        return u'%s' % self.pk


class Transcript(models.Model):
    name = models.CharField(max_length=45)
    strand = models.CharField(max_length=1)
    tss1500start = models.IntegerField()
    tss200start = models.IntegerField()
    txstart = models.IntegerField()
    txend = models.IntegerField()
    cdsstart = models.IntegerField()
    cdsend = models.IntegerField()
    description = models.TextField(max_length=3000)
    long_name = models.TextField(max_length=1000)
    has_coding = models.BooleanField(default=False)
    gene = models.ForeignKey(
        Gene, on_delete=models.CASCADE
    )

    def __str__(self):
        return '%s' % (self.name)

    class Meta:
        ordering = ['txstart']
        indexes = [
            models.Index(fields=['name']),
            models.Index(fields=['strand']),
            models.Index(fields=['tss1500start']),
            models.Index(fields=['tss200start']),
            models.Index(fields=['txstart']),
            models.Index(fields=['txend']),
            models.Index(fields=['cdsstart']),
            models.Index(fields=['cdsend']),
        ]
        index_together = [
            ["strand", "txstart", "txend", "gene"],
            ["strand", "tss1500start", "tss200start", "gene"],
        ]

    def __unicode__(self):
        return u'%s' % self.pk


class Feature(models.Model):
    feat_type = models.CharField(max_length=45)
    start = models.IntegerField()
    end = models.IntegerField()
    transcript = models.ForeignKey('Transcript', models.CASCADE)

    class Meta:
        ordering = ['start']
        unique_together = ("start", "end", "transcript")
        indexes = [
            models.Index(fields=['feat_type']),
        ]
        index_together = [
            ["start", "end"],
        ]



from django.contrib.auth.models import User

from dgenome.models import Genome, Chromosome, Gene, GeneLocation, GeneAlias
from dgenome.models import Transcript, Feature


def create_superuser(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return User.objects.create_superuser(**defaults)


def create_genome(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return Genome.objects.create(**defaults)


def create_chromosome(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return Chromosome.objects.create(**defaults)


def create_gene(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return Gene.objects.create(**defaults)


def create_genelocation(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return GeneLocation.objects.create(**defaults)


def create_genealias(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return GeneAlias.objects.create(**defaults)


def create_transcript(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return Transcript.objects.create(**defaults)


def create_feature(**kwargs):
    defaults = {}
    defaults.update(**kwargs)
    return Feature.objects.create(**defaults)



from django.urls import reverse
from django.test import TestCase

from rest_framework.test import APIClient

from .create_data import create_superuser
from .create_data import create_feature, create_transcript, create_genelocation
from .create_data import create_gene, create_genome, create_chromosome


class ViewTestCase(TestCase):

    def setUp(self):
        self.client = APIClient()
        self.user = create_superuser(username='admin',
                                     password='admin',
                                     email='admin@example.com')
        genome = create_genome(organism='human', chromosomes=22,
                               assembly='hg19', source='UCSC')
        chr = create_chromosome(name="chr1", genome=genome)
        gen = create_gene(name="Gen1", chromosome=chr)
        create_genelocation(txstart=1, txend=10, cdsstart=3, cdsend=7, gene=gen)
        create_genelocation(txstart=12, txend=22, cdsstart=14, cdsend=18, gene=gen)
        trans = create_transcript(name="Trans11", strand='-', tss1500start=1707,
                                  tss200start=207, txstart=1, txend=20, cdsstart=3,
                                  cdsend=7, gene=gen)
        create_feature(feat_type='exon', start=1, end=5, transcript=trans)
        create_feature(feat_type='intron', start=6, end=8, transcript=trans)
        create_feature(feat_type='exon', start=9, end=10, transcript=trans)

        gen = create_gene(name="Gen2", chromosome=chr)
        create_genelocation(txstart=6, txend=23, cdsstart=8, cdsend=18, gene=gen)
        trans = create_transcript(name="Trans21", strand='-', tss1500start=1707,
                                  tss200start=207, txstart=6, txend=23, cdsstart=8,
                                  cdsend=18, gene=gen)
        create_feature(feat_type='exon', start=6, end=9, transcript=trans)
        create_feature(feat_type='intron', start=10, end=15, transcript=trans)
        create_feature(feat_type='exon', start=16, end=23, transcript=trans)

    def test_health(self):
        response = self.client.get(reverse('dgenome:health'))
        self.assertEqual(response.status_code, 200)

    def test_chromosome_get(self):
        self.client.force_authenticate(self.user)
        response = self.client.get(reverse('dgenome:chromosome-list'),
                                   {'assembly': 'hg19'}, format='json')
        self.assertEqual(response.status_code, 200)
        for o in response.json():
            self.assertEqual(o['name'], 'chr1')

    def test_chromosome_name(self):
        self.client.force_authenticate(self.user)
        response = self.client.get(reverse('dgenome:chromosome-list'),
                                   {'assembly': 'hg19', 'name': 'chr1'}, format='json')
        self.assertEqual(response.status_code, 200)
        for o in response.json():
            self.assertEqual(o['name'], 'chr1')

    def test_gene_get_no_assembly(self):
        self.client.force_authenticate(self.user)
        response = self.client.get(reverse('dgenome:gene-list'),
                                   {'name': 'Gen1'}, format='json')
        self.assertEqual(response.status_code, 403)

    def test_gene_get(self):
        self.client.force_authenticate(self.user)
        response = self.client.get(reverse('dgenome:gene-list'),
                                   {'assembly': 'hg19', 'name': 'Gen1'}, format='json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        for o in response.json():
            self.assertEqual(o['name'], 'Gen1')

    def test_gene_get_by_pos(self):
        self.client.force_authenticate(self.user)
        response = self.client.get(reverse('dgenome:gene-list'),
                                   {'assembly': 'hg19', 'chr': 'chr1:4-12'}, format='json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 2)
        for o in response.json():
            f = False
            if o['name'] == 'Gen1':
                f = True
                self.assertEqual(len(o['genelocations']), 2)
            if o['name'] == 'Gen2':
                f = True
                self.assertEqual(len(o['genelocations']), 1)
            self.assertEqual(f, True)

    def test_transcripts_get(self):
        self.client.force_authenticate(self.user)
        response = self.client.get(reverse('dgenome:transcript-list'),
                                   {'assembly': 'hg19', 'name': 'Trans11'}, format='json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)
        for o in response.json():
            self.assertEqual(o['name'], 'Trans11')



from django.urls import path, include
from rest_framework import routers

from . import api
from dgenome.views import HealthCheck

router = routers.DefaultRouter()
router.register(r'genome', api.GenomeViewSet)
router.register(r'chromosome', api.ChromosomeViewSet)
router.register(r'gene', api.GeneViewSet)
router.register(r'transcript', api.TranscriptViewSet)

urlpatterns = (
    # urls for Django Rest Framework API
    path('api/v1/', include(router.urls)),
)

urlpatterns += (
    # urls for Chromosome
    path('health/', HealthCheck.as_view(), name='health'),
)



SECRET_KEY = '1978'

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'rest_framework',
    'dgenome'
]

ROOT_URLCONF = 'dgenome.tests.urls'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}



from . import models
from . import serializers
from .lib import utils

from rest_framework import viewsets, permissions
from rest_framework.exceptions import PermissionDenied


class GenomeViewSet(viewsets.ReadOnlyModelViewSet):
    """ViewSet for the Genome class"""
    queryset = models.Genome.objects.all()
    serializer_class = serializers.GenomeSerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        assembly = self.request.query_params.get('assembly', None)

        if assembly:
            return self.queryset.filter(assembly=assembly)

        return self.queryset


class ChromosomeViewSet(viewsets.ReadOnlyModelViewSet):
    """ViewSet for the Chromosome class"""

    queryset = models.Chromosome.objects.all()
    serializer_class = serializers.ChromosomeSerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        assembly = self.request.query_params.get('assembly', None)
        name = self.request.query_params.get('name', None)

        if not assembly:
            raise PermissionDenied()

        if name:
            return self.queryset.filter(name=name)

        return self.queryset


class GeneViewSet(viewsets.ReadOnlyModelViewSet):
    """ViewSet for the Gene class"""

    queryset = models.Gene.objects.all()
    serializer_class = serializers.GeneSerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        assembly = self.request.query_params.get('assembly', None)
        chr = self.request.query_params.get('chr', None)
        name = self.request.query_params.get('name', None)

        if not assembly:
            raise PermissionDenied()

        self.queryset = self.queryset.filter(chromosome__genome__assembly=assembly)

        if chr:
            return utils.get_genes(chr)

        if chr and name:
            return self.queryset.filter(chromosome__name=chr, name=name)
        elif chr and not name:
            self.queryset = self.queryset.filter(chromosome__name=chr)
        elif not chr and name:
            self.queryset = self.queryset.filter(name=name)

        return self.queryset


class TranscriptViewSet(viewsets.ReadOnlyModelViewSet):
    """ViewSet for the Transcript class"""

    queryset = models.Transcript.objects.all()
    serializer_class = serializers.TranscriptSerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        assembly = self.request.query_params.get('assembly', None)
        name = self.request.query_params.get('name', None)

        if not assembly:
            raise PermissionDenied()

        if name:
            self.queryset = self.queryset.filter(name=name)

        return self.queryset



from django.test import TestCase
from .create_data import create_feature, create_transcript, create_genelocation
from .create_data import create_gene, create_genome, create_chromosome

from dgenome.models import Chromosome, Gene, GeneLocation
from dgenome.models import Transcript


class ModelsTestCase(TestCase):
    def setUp(self):
        genome = create_genome(organism='human', chromosomes=22,
                               assembly='hg19', source='USCS')
        chr = create_chromosome(name="chr1", genome=genome)
        gen = create_gene(name="Gen1", chromosome=chr)
        create_genelocation(txstart=1, txend=10, cdsstart=3, cdsend=7, gene=gen)
        create_genelocation(txstart=11, txend=20, cdsstart=13, cdsend=17, gene=gen)
        trans = create_transcript(name="Trans1", strand='-', tss1500start=1707,
                                  tss200start=207, txstart=1, txend=20, cdsstart=3,
                                  cdsend=7, gene=gen)
        create_feature(feat_type='exon', start=1, end=5, transcript=trans)
        create_feature(feat_type='intron', start=6, end=8, transcript=trans)
        create_feature(feat_type='exon', start=9, end=10, transcript=trans)

    def test_chromosome_gene(self):
        chr = Chromosome.objects.get(name='chr1')
        self.assertEqual(str(chr), 'chr1')
        for g in chr.gene_set.all():
            self.assertEqual(g.name, 'Gen1')

    def test_chromosome_gene_relation(self):
        genes = Gene.objects.filter(chromosome__name='chr1')
        for g in genes:
            self.assertEqual(g.name, 'Gen1')

    def test_gene_chromosome(self):
        gen = Gene.objects.get(name="Gen1")
        self.assertEqual(gen.chromosome.name, 'chr1')
        self.assertEqual(str(gen), 'Gen1')

    def test_gene_location(self):
        gen = Gene.objects.get(name="Gen1")
        self.assertEqual(len(gen.genelocation_set.all()), 2)

    def test_location_gene(self):
        loc = GeneLocation.objects.get(txstart=11)
        self.assertEqual(loc.gene.name, 'Gen1')

    def test_transcript_gene(self):
        trans = Transcript.objects.get(txstart=1)
        self.assertEqual(trans.gene.name, 'Gen1')
        self.assertEqual(str(trans), 'Trans1')

    def test_gene_transcript(self):
        gen = Gene.objects.get(name="Gen1")
        for t in gen.transcript_set.all():
            self.assertEqual(t.name, 'Trans1')



from rest_framework import serializers

from dgenome.models import Genome, Chromosome, Gene, GeneLocation, GeneAlias, Transcript, Feature


class GenomeSerializer(serializers.ModelSerializer):
    class Meta:
        model = Genome
        fields = (
            'pk',
            'organism',
            'chromosomes',
            'assembly',
            'source',
        )


class ChromosomeSerializer(serializers.ModelSerializer):
    genome = GenomeSerializer()

    class Meta:
        model = Chromosome
        fields = '__all__'


class GeneLocationSerializer(serializers.ModelSerializer):
    class Meta:
        model = GeneLocation
        fields = '__all__'


class GeneAliasSerializer(serializers.ModelSerializer):
    class Meta:
        model = GeneAlias
        fields = '__all__'


class FeatureSerializer(serializers.ModelSerializer):
    class Meta:
        model = Feature
        fields = '__all__'


class TranscriptSerializer(serializers.ModelSerializer):
    features = FeatureSerializer(source="feature_set", many=True, read_only=True)

    class Meta:
        model = Transcript
        fields = '__all__'


class GeneSerializer(serializers.ModelSerializer):
    genelocations = GeneLocationSerializer(source="genelocation_set",
                                           many=True, read_only=True)
    genealias = GeneLocationSerializer(source="genealias_set",
                                       many=True, read_only=True)
    transcripts = TranscriptSerializer(source="transcript_set",
                                       many=True, read_only=True)
    chromosome = ChromosomeSerializer()

    class Meta:
        model = Gene
        fields = '__all__'



import re
from django.db.models import Q

from dgenome.models import Gene, Transcript, Feature


def get_coordinates_chr(term):
    chr_query = False
    start = False
    end = False
    m = re.search('chr(\\d+):(\\d+)(-(\\d+))?', term)
    if m:
        start = int(m.group(2))
        if m.group(4):
            end = int(m.group(4))
        if start <= 0:
            start = 1
        chr_query = 'chr' + m.group(1)
    return chr_query, start, end


def get_coordinates_gene(term):
    chr_query = False
    start = False
    end = False
    queryset = Gene.objects.filter(name=term.upper())
    if queryset:
        g = queryset.first()
        location = g.genelocation_set.first()
        start = location.start_pos
        end = location.end_pos
        chr_query = g.chr
    return chr_query, start, end


def get_coordinates(term):
    chr_query, start, end = get_coordinates_chr(term)
    return chr_query, start, end


def get_genes(term):
    chr_query, txstart, txend = get_coordinates(term)
    if chr_query and txstart and txstart:
        return Gene.objects.filter(Q(chromosome__name=chr_query),
                                   Q(genelocation__txend__lte=txend,
                                     genelocation__txend__gte=txstart) |
                                   Q(genelocation__txstart__lte=txend,
                                     genelocation__txstart__gte=txstart) |
                                   Q(genelocation__txstart__lte=txstart,
                                     genelocation__txend__gte=txend)).distinct()
    return Gene.objects.filter(name=term.upper())


def transcript_queryset_to_list(queryset):
    transcripts = []
    for t in queryset:
        transcripts.append(t.id)
    return transcripts


def find_tss1500(chr, mapinfo):
    queryset = Transcript.objects.filter(
        Q(strand='+', gene__chromosome__name=chr,
          tss1500start__lte=mapinfo, tss200start__gt=mapinfo) |
        Q(strand='-', gene__chromosome__name=chr,
          tss200start__lt=mapinfo, tss1500start__gte=mapinfo))
    return transcript_queryset_to_list(queryset)


def find_tss200(chr, mapinfo):
    queryset = Transcript.objects.filter(
        Q(strand='+', gene__chromosome__name=chr,
          tss200start__lte=mapinfo, txstart__gt=mapinfo) |
        Q(strand='-', gene__chromosome__name=chr,
          txend__lt=mapinfo, tss200start__gte=mapinfo))
    return transcript_queryset_to_list(queryset)


def find_body(chr, mapinfo):
    queryset = Transcript.objects.filter(gene__chromosome__name=chr,
                                         txstart__lte=mapinfo, txend__gte=mapinfo)
    return transcript_queryset_to_list(queryset)


def find_5utr(chr, mapinfo):
    queryset = Transcript.objects.filter(
        Q(strand='+', gene__chromosome__name=chr,
          txstart__lte=mapinfo, cdsstart__gt=mapinfo, has_coding=True) |
        Q(strand='-', gene__chromosome__name=chr,
          cdsend__lt=mapinfo, txend__gte=mapinfo, has_coding=True))
    return transcript_queryset_to_list(queryset)


def find_3utr(chr, mapinfo):
    queryset = Transcript.objects.filter(
        Q(strand='+', gene__chromosome__name=chr,
          cdsend__lt=mapinfo, txend__gte=mapinfo, has_coding=True) |
        Q(strand='-', gene__chromosome__name=chr,
          txstart__lte=mapinfo, cdsstart__gt=mapinfo, has_coding=True))
    return transcript_queryset_to_list(queryset)


def find_1exon(chr, mapinfo):
    transcripts = []
    feats = Feature.objects.filter(transcript__gene__chromosome__name=chr,
                                   feat_type='exon', start__lte=mapinfo,
                                   end__gte=mapinfo)
    for f in feats:
        if (f.transcript.strand == '+' and f.transcript.feature_set.first() == f) or (
                f.transcript.strand == '-' and f.transcript.feature_set.last() == f):
            transcripts.append(f.transcript.id)
    return transcripts



#!/usr/bin/env python

import sys
import json
import pandas
import django
import argparse

from django.core import management

django.setup()

from dgenome.models import Genome
from dgenome.models import Chromosome
from dgenome.models import Gene
from dgenome.models import Transcript
from dgenome.models import GeneLocation
from dgenome.models import Feature

from dgenome.serializers import ChromosomeSerializer
from dgenome.serializers import GeneSerializer
from dgenome.serializers import TranscriptSerializer
from dgenome.serializers import GeneLocationSerializer
from dgenome.serializers import FeatureSerializer


def insert_chromosomes(ref, genome, batch_size):
    print('Inserting Chromosome data')
    objects = []
    count = 0
    for c in ref[~ref[2].str.contains('_')][2].unique():
        objects.append(
            Chromosome(name=c, genome_id=genome.pk)
        )
        if len(objects) == batch_size:
            try:
                count += len(objects)
                print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count), end='\r')
                Chromosome.objects.bulk_create(objects)
                objects.clear()
            except:
                print('********************************************************************')
                for o in objects:
                    os = ChromosomeSerializer(o)
                    print(json.dumps(os.data, indent=4))
                    print('********************************************************************')
                sys.exit(-1)
    if objects:
        try:
            count += len(objects)
            print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count))
            Chromosome.objects.bulk_create(objects)
            objects.clear()
        except:
            print('********************************************************************')
            for o in objects:
                os = ChromosomeSerializer(o)
                print(json.dumps(os.data, indent=4))
                print('********************************************************************')
            sys.exit(-1)

    return {c.name: c.id for c in Chromosome.objects.all()}


def insert_genes(ref, chr_map, batch_size):
    print('Inserting Gene data')
    objects = []
    count = 0
    for c in chr_map:
        print('\nProcessing Chromosome: ' + c, end=' ')
        c_ref = ref[ref[2] == c][[12, 3]].drop_duplicates().sort_values(by=[12])
        cc_ref = c_ref.groupby(12, sort=False)[3].apply(''.join)
        cc_ref = cc_ref.to_frame()
        print(len(c_ref))
        for i, r in cc_ref.iterrows():
            objects.append(
                Gene(
                    name=i,
                    strand=r[3],
                    chromosome_id=chr_map[c]
                )
            )
            if len(objects) == batch_size:
                try:
                    count += len(objects)
                    print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count), end='\r')
                    Gene.objects.bulk_create(objects)
                    objects.clear()
                except:
                    print('********************************************************************')
                    for o in objects:
                        os = GeneSerializer(o)
                        print(json.dumps(os.data, indent=4))
                        print('********************************************************************')
                    sys.exit(-1)
    if objects:
        try:
            count += len(objects)
            print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count))
            Gene.objects.bulk_create(objects)
        except:
            print('********************************************************************')
            for o in objects:
                os = GeneSerializer(o)
                print(json.dumps(os.data, indent=4))
                print('********************************************************************')
            sys.exit(-1)

    return {g.chromosome.name + g.name: g.id for g in Gene.objects.all()}


def insert_transcripts(ref, chr_map, gene_map, batch_size):
    print('Inserting Transcript data')
    objects = []
    count = 0
    for c in chr_map:
        print('\nProcessing Chromosme: ' + c)
        c_ref = ref[ref[2] == c]
        for i, r in c_ref.iterrows():
            if r[3] == '+':
                tss200 = r[4] - 200
                tss1500 = tss200 - 1500
            else:
                tss200 = r[5] + 199
                tss1500 = tss200 + 1500
            cdsstart = r[6]
            cdsend = r[7]
            has_coding = False
            if cdsstart == cdsend:
                cdsstart -= 1
                cdsend -= 1
            else:
                cdsend -= 1
                has_coding = True
            objects.append(
                Transcript(
                    name=r[1],
                    strand=r[3],
                    tss1500start=tss1500,
                    tss200start=tss200,
                    txstart=r[4],
                    txend=r[5] - 1,
                    cdsstart=cdsstart,
                    cdsend=cdsend,
                    has_coding=has_coding,
                    gene_id=gene_map[c + r[12]]
                )
            )
            if len(objects) == batch_size:
                try:
                    count += len(objects)
                    print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count), end='\r')
                    Transcript.objects.bulk_create(objects)
                    objects.clear()
                except:
                    print('********************************************************************')
                    for o in objects:
                        os = TranscriptSerializer(o)
                        print(json.dumps(os.data, indent=4))
                        print('********************************************************************')
                    sys.exit(-1)
    if objects:
        try:
            count += len(objects)
            print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count))
            Transcript.objects.bulk_create(objects)
        except:
            print('********************************************************************')
            for o in objects:
                os = TranscriptSerializer(o)
                print(json.dumps(os.data, indent=4))
                print('********************************************************************')
            sys.exit(-1)

    return {t.name: t.id for t in Transcript.objects.all()}


def insert_features(ref, chr_map, trasc_map, batch_size):
    print('Inserting Feature data')
    objects = []
    count = 0
    for c in chr_map:
        print('\nProcessing Chromosme: ' + c)
        c_ref = ref[ref[2] == c]
        for i, r in c_ref.iterrows():
            exon_start = [int(l) for l in r[9].split(',') if l]
            exon_end = [int(l) - 1 for l in r[10].split(',') if l]
            if len(exon_start) != len(exon_end):
                raise ValueError('Different number of exon start and end coordinates')
            for i in range(len(exon_start)):
                objects.append(
                    Feature(
                        feat_type='exon',
                        start=exon_start[i],
                        end=exon_end[i],
                        transcript_id=trasc_map[r[1]]
                    )
                )
                if i < len(exon_start) - 1:
                    start = exon_end[i] + 1
                    end = exon_start[i + 1] - 1
                    objects.append(
                        Feature(
                            feat_type='intron',
                            start=start,
                            end=end,
                            transcript_id=trasc_map[r[1]]
                        )
                    )
                if len(objects) >= batch_size:
                    try:
                        count += len(objects)
                        print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count), end='\r')
                        Feature.objects.bulk_create(objects)
                        objects.clear()
                    except:
                        print('********************************************************************')
                        for o in objects:
                            os = FeatureSerializer(o)
                            print(json.dumps(os.data, indent=4))
                            print('********************************************************************')
                        sys.exit(-1)
    if objects:
        try:
            count += len(objects)
            print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count))
            Feature.objects.bulk_create(objects)
            objects.clear()
        except:
            print('********************************************************************')
            for o in objects:
                os = FeatureSerializer(o)
                print(json.dumps(os.data, indent=4))
                print('********************************************************************')
            sys.exit(-1)


def insert_gene_location(batch_size):
    print('Inserting Gene Location data')
    objects = []
    count = 0
    for g in Gene.objects.all():
        loc = []
        for t in g.transcript_set.all():
            if not loc:
                loc.append({'txstart': t.txstart, 'txend': t.txend,
                            'cdsstart': t.cdsstart, 'cdsend': t.cdsend})
            else:
                append = False
                for l in loc:
                    if t.txend < l['txstart'] or t.txstart > l['txend']:
                        append = True
                    else:
                        l['txstart'] = min(l['txstart'], t.txstart)
                        l['txend'] = max(l['txend'], t.txend)
                        l['cdsstart'] = min(l['cdsstart'], t.cdsstart)
                        l['cdsend'] = max(l['cdsend'], t.cdsend)
                        append = False
                        break
                if append:
                    loc.append({'txstart': t.txstart, 'txend': t.txend,
                                'cdsstart': t.cdsstart, 'cdsend': t.cdsend})
        for l in loc:
            objects.append(
                GeneLocation(
                    txstart=l['txstart'],
                    txend=l['txend'],
                    cdsstart=l['cdsstart'],
                    cdsend=l['cdsend'],
                    gene_id=g.pk
                )
            )
            if len(objects) == batch_size:
                try:
                    count += len(objects)
                    print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count), end='\r')
                    GeneLocation.objects.bulk_create(objects)
                    objects.clear()
                except:
                    print('********************************************************************')
                    for o in objects:
                        os = GeneLocationSerializer(o)
                        print(json.dumps(os.data, indent=4))
                        print('********************************************************************')
                    sys.exit(-1)
    if objects:
        try:
            count += len(objects)
            print('Inserting ' + str(len(objects)) + ' Total inserted: ' + str(count))
            GeneLocation.objects.bulk_create(objects)
        except:
            print('********************************************************************')
            for o in objects:
                os = GeneLocationSerializer(o)
                print(json.dumps(os.data, indent=4))
                print('********************************************************************')
            sys.exit(-1)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Parse and insert the UCSC GTF file in the database')
    parser.add_argument('-r', help='UCSC refGene.txt', required=True)
    parser.add_argument('-b', help='Batch size', required=True)
    parser.add_argument('-o', help='Organism', required=True)
    parser.add_argument('-c', help='Number of Chromosomes', required=True)
    parser.add_argument('-a', help='Assembly', required=True)
    parser.add_argument('-s', help='Genome source', required=True)

    args = parser.parse_args()
    ref_file = args.r
    batch_size = int(args.b)
    org = args.o
    chrs = int(args.c)
    assembly = args.a
    source = args.s

    management.call_command("makemigrations", "dgenome")
    management.call_command("migrate", "dgenome")

    genome = Genome.objects.create(organism=org, chromosomes=chrs, assembly=assembly, source=source)
    genome.save()

    print('Loading ref')
    ref = pandas.read_csv(ref_file, sep='\t', header=None)

    chr_map = insert_chromosomes(ref, genome, batch_size)
    gene_map = insert_genes(ref, chr_map, batch_size)
    trasc_map = insert_transcripts(ref, chr_map, gene_map, batch_size)
    insert_features(ref, chr_map, trasc_map, batch_size)
    insert_gene_location(batch_size)

    print('Genes: ' + str(Gene.objects.all().count()) + ' Locations: ' + str(GeneLocation.objects.all().count()))
    print('Transcripts: ' + str(Transcript.objects.all().count()))
    print('Features: ' + str(Feature.objects.all().count()))



from django.http import HttpResponse
from django.views.generic.base import View


class HealthCheck(View):
    """
    Application health check.

    This view is used by robots to check if the application is in good shape.
    All the necessary backend checks can be implemented here as well.
    """
    http_method_names = ['get']

    def get(self, request):
        return HttpResponse('The application is healthy')
